/////////////////////////////////////////////////////////////////////
// Class Creator Version 2.0.000 Copyrigth (C) Poul A. Costinsky 1994
///////////////////////////////////////////////////////////////////
// Header File WizSortedArr.h
// class CWizSortedArr
// Class derived from CArray wich always keeps elements sorted.
// Needs template argument COMPARE_TYPE, which compares two
// elements and returns negative, 0, or positive.
// Requirement:
// class Comparer:
// {
// public:
//		int Compare (const TYPE& a, const TYPE& b) const;
// };
// The array never duplicates two equal elements.
//
//
// 05/05/1996 13:32                             Author: Poul
//
///////////////////////////////////////////////////////////////////

#ifndef __CWizSortedArr_H
#define __CWizSortedArr_H

#include "WizArr.h"

/////////////////////////////////////////////////////////////////////////////
// class CWizSortedArr
template <class TYPE, class COMPARE_TYPE>
class CWizSortedArr : protected CWizArr<TYPE,TYPE&>
{
public:
// Constructors:
	CWizSortedArr () {}
// Destructor:
	virtual ~CWizSortedArr () {}
public:
// Operations:
	const TYPE& operator[](int i) const 
		{ return const_cast<CWizSortedArr*>(this)->ElementAt(i); }
	int		Add(const TYPE& elem);	// Adds sorted and returns 
									//index of added or 
									// existing element.
	int		Find(const TYPE& elem) const; // Finds element and returns
									// index or -1 if not exist
	int		Delete(const TYPE& elem); // deletes element and returns its
									// index or -1 if not exist
	int		Replace(const TYPE& elem1, const TYPE& elem2); // Replaces elem1 with elem2
	int		GetSize() const { return CWizArr<TYPE,TYPE&>::GetSize(); }
public:
	struct Found {
		const int	nIndex;
		const BOOL	bExist;
		Found (int i, BOOL b)
			: nIndex(i), bExist(b) { ASSERT(nIndex >=0); }
		};
protected:	
// Implementation:
	Found FindElement (const TYPE& elem) const;
};


/////////////////////////////////////////////////////////////////////////////
//*****************************************************************
template <class TYPE, class COMPARE_TYPE>
inline int CWizSortedArr<TYPE, COMPARE_TYPE>::Add (const TYPE& elem)
{
	Found f = FindElement(elem);
	if (!f.bExist)
		InsertAt(f.nIndex, (TYPE&)elem, 1);

	return f.nIndex;
}
//*****************************************************************
template <class TYPE, class COMPARE_TYPE>
inline int CWizSortedArr<TYPE, COMPARE_TYPE>::Delete (const TYPE& elem)
{
	Found f = FindElement(elem);
	if (f.bExist)
		{
		RemoveAt(f.nIndex);
		return f.nIndex;
		}
	return -1;
}
//*****************************************************************
template <class TYPE, class COMPARE_TYPE>
inline int CWizSortedArr<TYPE, COMPARE_TYPE>::Replace(const TYPE& elem1, const TYPE& elem2)
{
	Delete(elem1);
	return Add(elem2);
}
//*****************************************************************
template <class TYPE, class COMPARE_TYPE>
inline int CWizSortedArr<TYPE, COMPARE_TYPE>::Find (const TYPE& elem) const
{
	Found f = FindElement(elem);
	return (f.bExist) ? f.nIndex : -1;
}
//*****************************************************************
template <class TYPE, class COMPARE_TYPE>
CWizSortedArr<TYPE, COMPARE_TYPE>::Found 
CWizSortedArr<TYPE, COMPARE_TYPE>::FindElement (const TYPE& elem) const
{
	INDEX nLast  = GetSize();
	if (nLast == 0)
		return Found(0, FALSE);
	COMPARE_TYPE comparer;
	int res = comparer.Compare (elem, (*this)[0]);
	if (res <= 0)
		return Found(0, (res == 0));
	res = comparer.Compare (elem, (*this)[nLast-1]);
	if (res == 0)
		return Found(nLast-1, TRUE);
	if (res > 0)
		return Found(nLast, FALSE);

	INDEX nFirst = 0;
	while (nLast > nFirst)
		{
		int nMiddle = (nLast + nFirst)/2;
		int res = comparer.Compare (elem, (*this)[nMiddle]);
		if (res == 0)
			return Found(nMiddle, TRUE);

		if (res < 0)
			{
			if (nFirst >= nMiddle - 1)
				{
				if (nFirst != nMiddle)
					res = comparer.Compare (elem, (*this)[nFirst]);
				if (res == 0)
					return Found(nFirst, TRUE);
				else if (res < 0)
					return Found(nFirst, FALSE);
				else
					return Found(nFirst + 1, FALSE);
				}
			nLast = nMiddle;
			}
		else
			{
			if (nLast <= nMiddle + 1)
				{
				if (nLast != nMiddle + 1)
					res = comparer.Compare (elem, (*this)[nLast - 1]);
				if (res == 0)
					return Found(nLast - 1, TRUE);
				else if (res > 0)
					return Found(nLast, FALSE);
				else
					return Found(nLast, FALSE);
				}
			nFirst = nMiddle;
			}
		}
	ASSERT(0);
	return Found(nFirst, FALSE);
}
//*****************************************************************
/////////////////////////////////////////////////////////////////////////////
#endif // __CWizSortedArr_H


